import logging
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)

from scapy.all   import *
from scapy.utils import PcapWriter

# test source port

packet = Ether()/IP()/TCP(sport=8080)

cap = PcapWriter("pcaps/test_sport.pcap")
cap.write(packet)

# test destination port

packet = Ether()/IP()/TCP(dport=9999)

cap = PcapWriter("pcaps/test_dport.pcap")
cap.write(packet)

# test sequence number

packet = Ether()/IP()/TCP(seq=1200)

cap = PcapWriter("pcaps/test_seq.pcap")
cap.write(packet)

# test acknowledgement number

packet = Ether()/IP()/TCP(ack=5000, flags='A')

cap = PcapWriter("pcaps/test_ack.pcap")
cap.write(packet)

# test cwr

packet = Ether()/IP()/TCP(flags='C')

cap = PcapWriter("pcaps/test_cwr.pcap")
cap.write(packet)

# test ece

packet = Ether()/IP()/TCP(flags='E')

cap = PcapWriter("pcaps/test_ece.pcap")
cap.write(packet)

# test urg

packet = Ether()/IP()/TCP(flags='U')

cap = PcapWriter("pcaps/test_urg.pcap")
cap.write(packet)

# test ack
# tested by ack num above

# test psh

packet = Ether()/IP()/TCP(flags='P')

cap = PcapWriter("pcaps/test_psh.pcap")
cap.write(packet)

# test rst

packet = Ether()/IP()/TCP(flags='R')

cap = PcapWriter("pcaps/test_rst.pcap")
cap.write(packet)

# test syn

packet = Ether()/IP()/TCP(flags='S')

cap = PcapWriter("pcaps/test_syn.pcap")
cap.write(packet)

# test fin

packet = Ether()/IP()/TCP(flags='F')

cap = PcapWriter("pcaps/test_fin.pcap")
cap.write(packet)

# test window size

packet = Ether()/IP()/TCP(window=2000)

cap = PcapWriter("pcaps/test_window.pcap")
cap.write(packet)

# test checksum

packet = Ether()/IP()/TCP(chksum=15)

cap = PcapWriter("pcaps/test_chksum.pcap")
cap.write(packet)

# test urgent pointer

packet = Ether()/IP()/TCP(urgptr=10)

cap = PcapWriter("pcaps/test_urgptr.pcap")
cap.write(packet)

# test eol option

packet = Ether()/IP()/TCP(options=[('EOL', None)])

cap = PcapWriter("pcaps/test_eolop.pcap")
cap.write(packet)

# test noop option

packet = Ether()/IP()/TCP(options=[('NOP', None)])

cap = PcapWriter("pcaps/test_noopop.pcap")
cap.write(packet)

# test mss option

packet = Ether()/IP()/TCP(options=[('MSS', 1200)])

cap = PcapWriter("pcaps/test_mssop.pcap")
cap.write(packet)

# test ts option

packet = Ether()/IP()/TCP(options=[('Timestamp', (20, 10))])

cap = PcapWriter("pcaps/test_tsop.pcap")
cap.write(packet)

# test sack permitted option

packet = Ether()/IP()/TCP(options=[('SAckOK', '')])

cap = PcapWriter("pcaps/test_sackpermop.pcap")
cap.write(packet)

# test sack block option

packet = Ether()/IP()/TCP(options=[('SAck', (1000, 2000))])

cap = PcapWriter("pcaps/test_sackop.pcap")
cap.write(packet)

# test payload

packet = Ether()/IP()/TCP()/Raw(load="Hello, world.")

cap = PcapWriter("pcaps/test_payload.pcap")
cap.write(packet)

# test invalid option (kind)

packet = Ether()/IP()/TCP(options=[(34, b'')])

cap = PcapWriter("pcaps/test_badkind.pcap")
cap.write(packet)

# test invalid data offset size

packet = Ether()/IP()/TCP(dataofs=1)

cap = PcapWriter("pcaps/test_badoffset.pcap")
cap.write(packet)

# test syn and fin being set at the same time

packet = Ether()/IP()/TCP(flags='SF')

cap = PcapWriter("pcaps/test_synfin.pcap")
cap.write(packet)

# test syn and fin being set at the same time

packet = Ether()/IP()/TCP(reserved=12)

cap = PcapWriter("pcaps/test_nonzerorsvd.pcap")
cap.write(packet)
